@@ -1,5 +1,7 @@
+from collections import namedtuple
 from functools import partial
 from itertools import groupby
+from math import ceil
 from struct import Struct, unpack
 from sqlite3 import connect
 
@@ -8,6 +10,7 @@ def stream_sqlite(sqlite_chunks):
     LEAF_INDEX = b'\x0a'
     LEAF_TABLE = b'\x0d'
 
+    unsigned_char = Struct('B')
     unsigned_short = Struct('>H')
     unsigned_long = Struct('>L')
     table_leaf_header = Struct('>HHHB')
@@ -15,6 +18,8 @@ def stream_sqlite(sqlite_chunks):
     index_interior_header = Struct('>HHHBL')
     freelist_trunk_header = Struct('>LL')
 
+    column_constructor = namedtuple('Column', ('cid', 'name', 'type', 'notnull', 'dflt_value', 'pk'))
+
     def get_byte_reader(iterable):
         chunk = b''
         offset = 0
@@ -82,24 +87,41 @@ def stream_sqlite(sqlite_chunks):
         if encoding not in (0, 1):
             raise ValueError('Unsupported encoding')
 
+        reserved_space, = unsigned_char.unpack(header[20:21])
+        if reserved_space != 0:
+            raise ValueError('Reserved space is not supported')
+
         page_size, = unsigned_short.unpack(header[16:18])
         page_size = 65536 if page_size == 1 else page_size
         num_pages_expected, = unsigned_long.unpack(header[28:32])
         first_freelist_trunk_page, = unsigned_long.unpack(header[32:36])
+        incremental_vacuum = any(header[52:56])
 
-        return page_size, num_pages_expected, first_freelist_trunk_page
+        return page_size, num_pages_expected, first_freelist_trunk_page, incremental_vacuum
 
-    def yield_page_nums_pages_readers(get_bytes, page_size, num_pages_expected):
+    def yield_page_nums_pages_readers(get_bytes, page_size, num_pages_expected, incremental_vacuum):
         page_bytes = bytes(100) + get_bytes(page_size - 100)
         page_reader, _ = get_chunk_readers(page_bytes)
         page_reader(100)
 
+        lock_byte_page = 1073741824 // page_size + 1
+        ptrmap_j = int(ceil(page_size/5))
+
         yield 1, page_bytes, page_reader
 
         for page_num in range(2, num_pages_expected + 1):
             page_bytes = get_bytes(page_size)
             page_reader, _ = get_chunk_readers(page_bytes)
 
+            # ptrmap and lock bytes pages are not processed in any way
+            ptrmap_page_prev = incremental_vacuum and (page_num - 1 - 2) % ptrmap_j == 0
+            ptrmap_page_curr = incremental_vacuum and (page_num - 2) % ptrmap_j == 0
+            lock_byte_page_prev = (page_num - 1) == lock_byte_page
+            lock_byte_page_curr = page_num == lock_byte_page
+
+            if ptrmap_page_curr or lock_byte_page_curr or (ptrmap_page_prev and lock_byte_page_prev):
+                continue
+
             yield page_num, page_bytes, page_reader
 
     def yield_table_pages(page_nums_pages_readers, first_freelist_trunk_page):
@@ -111,26 +133,26 @@ def stream_sqlite(sqlite_chunks):
         # identify a page, but don't have it in the buffer yet.
         page_processors = {}
 
-        def process_table_page(table_name, table_info, page_bytes, page_reader):
+        def process_table_page(table_name, table_info, row_constructor, page_bytes, page_reader):
 
             def get_master_table(master_cells):
 
-                def schema(cur, table_name, sql):
+                def table_info_and_row_constructor(cur, table_name, sql):
                     cur.execute(sql)
                     cur.execute("PRAGMA table_info('" + table_name.replace("'","''") + "');")
-                    rows = cur.fetchall()
-                    cols = [d[0] for d in cur.description]
-                    return tuple({col: row[i] for i, col in enumerate(cols)} for row in rows)
+                    columns = cur.fetchall()
+                    row_constructor = namedtuple('Row', (column[1] for column in columns))
+                    return tuple(column_constructor(*column) for column in columns), row_constructor
 
                 with connect(':memory:') as con:
                     cur = con.cursor()
 
                     return tuple(
                         (
-                            cell[0],                                                        # table or index
-                            cell[1],                                                        # table name
-                            schema(cur, cell[1], cell[4]) if cell[0] == 'table' else None,  # table info
-                            cell[3],                                                        # root page
+                            cell[0],  # table or index
+                            cell[1],  # table name
+                            table_info_and_row_constructor(cur, cell[1], cell[4]) if cell[0] == 'table' else (None, None),
+                            cell[3],  # root page
                         )
                         for cell in master_cells
                         if cell[0] in ('table', 'index')
@@ -177,9 +199,9 @@ def stream_sqlite(sqlite_chunks):
                     table_leaf_header.unpack(page_reader(7))
 
                 pointers = unpack('>{}H'.format(num_cells), page_reader(num_cells * 2))
-                for table_or_index, table_name, table_info, root_page in get_master_table(yield_leaf_table_cells(pointers)):
+                for table_or_index, table_name, (table_info, row_constructor), root_page in get_master_table(yield_leaf_table_cells(pointers)):
                     yield from (
-                        process_if_buffered_or_remember(partial(process_table_page, table_name, table_info), root_page) if table_or_index == 'table' else \
+                        process_if_buffered_or_remember(partial(process_table_page, table_name, table_info, row_constructor), root_page) if table_or_index == 'table' else \
                         process_if_buffered_or_remember(process_index_page, root_page)
                     )
 
@@ -190,10 +212,7 @@ def stream_sqlite(sqlite_chunks):
                 pointers = unpack('>{}H'.format(num_cells), page_reader(num_cells * 2))
 
                 yield table_name, table_info, (
-                    {
-                        table_info[i]['name']: value
-                        for i, value in enumerate(cell)
-                    }
+                    row_constructor(*cell)
                     for cell in yield_leaf_table_cells(pointers)
                 )
 
@@ -207,10 +226,10 @@ def stream_sqlite(sqlite_chunks):
                     cell_num_reader, cell_varint_reader = get_chunk_readers(page_bytes, pointer)
                     page_number, =  unsigned_long.unpack(cell_num_reader(4))
                     yield from process_if_buffered_or_remember(
-                        partial(process_table_page, table_name, table_info), page_number)
+                        partial(process_table_page, table_name, table_info, row_constructor), page_number)
 
                 yield from process_if_buffered_or_remember(
-                    partial(process_table_page, table_name, table_info), right_most_pointer)
+                    partial(process_table_page, table_name, table_info, row_constructor), right_most_pointer)
 
             page_type = page_reader(1)
             yield from (
@@ -264,7 +283,7 @@ def stream_sqlite(sqlite_chunks):
             else:
                 yield from process(page_bytes, page_reader)
 
-        page_processors[1] = partial(process_table_page, 'sqlite_schema', ())
+        page_processors[1] = partial(process_table_page, 'sqlite_schema', (), column_constructor)
         page_processors[first_freelist_trunk_page] = process_freelist_trunk_page
 
         for page_num, page_bytes, page_reader in page_nums_pages_readers:
@@ -290,8 +309,8 @@ def stream_sqlite(sqlite_chunks):
             yield name, info, _rows(single_table_pages)
 
     get_bytes = get_byte_reader(sqlite_chunks)
-    page_size, num_pages_expected, first_freelist_trunk_page = parse_header(get_bytes(100))
+    page_size, num_pages_expected, first_freelist_trunk_page, incremental_vacuum = parse_header(get_bytes(100))
 
-    page_nums_pages_readers = yield_page_nums_pages_readers(get_bytes, page_size, num_pages_expected)
+    page_nums_pages_readers = yield_page_nums_pages_readers(get_bytes, page_size, num_pages_expected, incremental_vacuum)
     table_pages = yield_table_pages(page_nums_pages_readers, first_freelist_trunk_page)
     yield from group_by_table(table_pages)
